#include <windows.h> 
#include <assert.h>
#include "game.h"
#include "main.h"

/* global variables */
int numSuitsToUse = 2;
Card deck[NUM_CARDS_IN_STOCK];
CardList *tableau[NUM_TABLEAU_SLOTS];
int remainingCards;
int finished_piles[8];
Undo *undo = NULL;

/* function prototypes */
static CardList *cardlist_insert(CardList *list, Card c);
static CardList *cardlist_add(CardList *list, Card c);
static int myrand(int n);
static void remove_finished_sequence(int slot, CardList *cl);
static void replace_finished_sequence(int finished_pile, int slot);
static void undeal_more_cards(void);
static void unreveal_top_card(int slot);

void
checkForCompletedPiles(int slot)
{
	assert(slot >= 0 && slot < NUM_TABLEAU_SLOTS);

	CardList *cl = tableau[slot];
	CardList *startCL = NULL;

	while (cl->next != NULL)
	{
		if (cl->card.rank == 12 /* highest rank */
		      && cl->card.face_up)
		{
			/* this might be the beginning of a finished sequence */
			startCL = cl;
		}

		if (startCL != NULL)
		{
			/* check whether next card is same suit/rank */
			if (!(cl->card.rank == cl->next->card.rank + 1
			       && cl->card.suit == cl->next->card.suit
			       && cl->next->card.face_up))
			{
				/* it's not, so cancel this sequence */
				startCL = NULL;
			}
			else if (cl->next->card.rank == 0)
			{
				/* a finished sequence */
				remove_finished_sequence(slot, startCL);
				reveal_top_card(slot);
				return;
			}
		}
		cl = cl->next;
	}

	/* down here means no sequences were found */
	return;
}

void
cleanup_game(void)
{
	int i;

	for (i = 0; i < NUM_TABLEAU_SLOTS; i++)
	{
		tableau[i] = cardlist_free(tableau[i]);
	}

	/* take care of undo information */
	while (undo != NULL)
	{
		Undo *next = undo->next;
		free(undo);
		undo = next;
	}
}

void
deal_more_cards(void)
{
	int i;

	for (i = 0; i < NUM_TABLEAU_SLOTS; i++)
	{
		if (remainingCards > 0)
		{
			deck[remainingCards - 1].face_up = TRUE;
			tableau[i] = cardlist_add(tableau[i], deck[--remainingCards]);
		}
	}
	main_invalidate_all();

	Undo *new_undo = calloc(1, sizeof(Undo));
	new_undo->type = UNDO_DEAL;
	new_undo->next = undo;
	undo = new_undo;
}

int
game_get_completed_count(void)
{
	int count = 0;
	int i;
	for (i = 0; i < 8; i++)
	{
		if (finished_piles[i])
			count++;
	}
	return count;
}

CardList *
getCardListAtPosition(int col, int row)
{
	if (col < 0 || col >= NUM_TABLEAU_SLOTS)
		return NULL;

	CardList *p = tableau[col];
	while (p != NULL && row > 0)
	{
		p = p->next;
		row--;
	}

	return p;
}

int
get_suit_of_finished_pile(int pile)
{
	assert(pile >= 0 && pile < 8);
	assert(numSuitsToUse == 1 || numSuitsToUse == 2 || numSuitsToUse == 4);
	return 3 - pile % numSuitsToUse;
}

void
moveCards(int fromCol, int fromRow, int destCol, int bSaveUndo)
{
	CardList *srcCL;
	CardList *destCL;

	if (bSaveUndo)
	{
		int destRow = cardlist_len(tableau[destCol]);
		Undo *new_undo = calloc(1, sizeof(Undo));
		new_undo->type = UNDO_MOVE;
		new_undo->fromSlot = destCol;
		new_undo->toSlot = fromCol;
		new_undo->cardNum = destRow;
		new_undo->next = undo;
		undo = new_undo;
	}

	srcCL = getCardListAtPosition(fromCol, fromRow);
	assert(srcCL != NULL);

	assert(destCol >= 0 && destCol < NUM_TABLEAU_SLOTS);
	destCL = tableau[destCol];
	if (destCL == NULL)
	{
		/* move to an empty slot */
		/* - detach cards from source */
		tableau[fromCol] = cardlist_detach(tableau[fromCol], srcCL);

		/* - attach to destination */
		assert(tableau[destCol] == NULL);
		tableau[destCol] = srcCL;
	}
	else
	{
		assert(destCL != NULL);

		/* find last card in slot */
		while (destCL->next != NULL)
			destCL = destCL->next;

		/* move the cards */
		/* - detach cards from source */
		tableau[fromCol] = cardlist_detach(tableau[fromCol], srcCL);

		/* - attach to destination */
		assert(destCL->next == NULL);
		destCL->next = srcCL;
	}

	/* maybe turn up the top card on the source stack */
	reveal_top_card(fromCol);

	/* maybe this move caused a sequence of 13 to be completed */
	checkForCompletedPiles(destCol);

	/* redraw the affected card piles */
	main_invalidate_card_rect(fromCol, 0, 0);
	main_invalidate_card_rect(destCol, 0, 0);
}

void
game_shuffle(void)
{
	int i;

	/* construct the deck */
	for (i = 0; i < NUM_CARDS_IN_STOCK; i++)
	{
		deck[i].suit = 3 - (i / 13) % numSuitsToUse;
		deck[i].rank = i % 13;
		deck[i].face_up = 0;
	}

	/* shuffle the deck */
	for (i = 0; i < NUM_CARDS_IN_STOCK; i++)
	{
		int j = myrand(NUM_CARDS_IN_STOCK - i);
		Card c = deck[i];
		deck[i] = deck[j];
		deck[j] = c;
	}
}

void
start_game(void)
{
	int i;

	/* all cards are in the deck */
	remainingCards = NUM_CARDS_IN_STOCK;

	/* put cards in tableau */
	for (i = 0; i < NUM_TABLEAU_SLOTS; i++)
	{
		int j;
		tableau[i] = cardlist_free(tableau[i]);
		for (j = 0; j < ((i % 3 == 0) ? 6 : 5); j++)
		{
			Card c = deck[--remainingCards];
			c.face_up = (j == 0);
			tableau[i] = cardlist_insert(tableau[i], c);
		}
	}

	/* clear foundation */
	for (i = 0; i < 8; i++)
	{
		finished_piles[i] = FALSE;
	}
	main_invalidate_all();
}

static void
remove_finished_sequence(int slot, CardList *cl)
{
	int suit = cl->card.suit;

	/* remove cards from board */
	tableau[slot] = cardlist_detach(tableau[slot], cl);
	cardlist_free(cl);

	/* add to finished piles */
	int i;
	for (i = 0; i < 8; i++)
	{
		if (!finished_piles[i] && get_suit_of_finished_pile(i) == suit)
		{
			finished_piles[i] = TRUE;
			break;
		}
	}
	assert(i < 8);

	/* create Undo information */
	Undo *new_undo = calloc(1, sizeof(Undo));
	new_undo->type = UNDO_FINISH;
	new_undo->fromSlot = slot;
	new_undo->toSlot = i;
	new_undo->next = undo;
	undo = new_undo;

	main_invalidate_status_area();
	main_invalidate_card_column(slot);
}

static void
replace_finished_sequence(int finished_pile, int slot)
{
	assert(finished_pile >= 0 && finished_pile < 8);
	assert(slot >= 0 && slot < NUM_TABLEAU_SLOTS);

	int suit = get_suit_of_finished_pile(finished_pile);
	finished_piles[finished_pile] = FALSE;

	int r;
	for (r = 13 - 1; r >= 0; r--)
	{
		Card c;
		c.suit = suit;
		c.rank = r;
		c.face_up = TRUE;
		tableau[slot] = cardlist_add(tableau[slot], c);
	}

	main_invalidate_status_area();
	main_invalidate_card_column(slot);
}

/**
 * Turns the top card of the specified tableau slot face-up.
 */
void
reveal_top_card(int slot)
{
	assert(slot >= 0 && slot < NUM_TABLEAU_SLOTS);

	CardList *cl = tableau[slot];
	if (cl == NULL)
		return;
	/* find top card in pile */
	while (cl->next != NULL)
		cl = cl->next;

	if (!cl->card.face_up)
	{
		cl->card.face_up = TRUE;
		main_invalidate_card_column(slot);

		/* record "undo" information */
		Undo *new_undo = calloc(1, sizeof(Undo));
		new_undo->type = UNDO_REVEAL;
		new_undo->toSlot = slot;
		new_undo->next = undo;
		undo = new_undo;
	}
}

static void
undeal_more_cards(void)
{
	int i;

	for (i = 0; i < NUM_TABLEAU_SLOTS; i++)
	{
		/* find last card */
		CardList *cl = tableau[i];
		while (cl != NULL && cl->next != NULL)
			cl = cl->next;

		/* remove the last card */
		if (cl != NULL)
		{
			tableau[i] = cardlist_detach(tableau[i], cl);
			cardlist_free(cl);
			remainingCards++;
		}
	}
	main_invalidate_all();
}

void
undo_last_move(void)
{
	if (undo == NULL)
		return;

	Undo *next_undo = undo->next;
	BOOL auto_undo = FALSE;

	switch (undo->type)
	{
	case UNDO_MOVE:
		moveCards(undo->fromSlot, undo->cardNum, undo->toSlot, FALSE);
		break;

	case UNDO_REVEAL:
		unreveal_top_card(undo->toSlot);
		auto_undo = TRUE;
		break;

	case UNDO_FINISH:
		replace_finished_sequence(undo->toSlot, undo->fromSlot);
		auto_undo = TRUE;
		break;
	
	case UNDO_DEAL:
		undeal_more_cards();
		break;
	}

	free(undo);
	undo = next_undo;

	if (auto_undo)
		undo_last_move();
}

static void
unreveal_top_card(int slot)
{
	assert(slot >= 0 && slot < NUM_TABLEAU_SLOTS);

	CardList *cl = tableau[slot];
	if (cl == NULL)
		return;
	/* find top card in pile */
	while (cl->next != NULL)
		cl = cl->next;

	if (cl->card.face_up)
	{
		cl->card.face_up = FALSE;
		main_invalidate_card_column(slot);
	}
}

CardList *
cardlist_free(CardList *list)
{
	if (list != NULL)
	{
		cardlist_free(list->next);
	}
	free(list);
	return NULL;
}

static CardList *
cardlist_insert(CardList *list, Card c)
{
	CardList *nl;

	nl = calloc(1, sizeof(CardList));
	nl->card = c;
	nl->next = list;
	return nl;
}

/**
 * Adds a card to the _end_ of the specified linked list.
 * Verified working as of 2008-05-28.
 *
 * @return a pointer to the head of the list
 */
static CardList *
cardlist_add(CardList *list, Card c)
{
	CardList *nl;

	nl = calloc(1, sizeof(CardList));
	nl->card = c;
	nl->next = NULL;

	if (list == NULL)
	{
		return nl;
	}
	else
	{
		/* find last entry of the given list */
		CardList *p = list;
		assert(p != NULL);
		while (p->next != NULL)
		{
			p = p->next;
		}
		p->next = nl;
		return list;
	}
}

int
cardlist_len(const CardList *list)
{
	int len = 0;
	while (list != NULL)
	{
		len++;
		list = list->next;
	}
	return len;
}

CardList *
cardlist_detach(CardList *list, CardList *part)
{
	if (list == part)
		return NULL;

	CardList *cl = list;
	while (cl != NULL)
	{
		if (cl->next == part)
		{
			cl->next = NULL;
			break;
		}
		cl = cl->next;
	}
	return list;
}

static int
myrand(int n)
{
	return rand() % n;
}
