#include "main.h"
#include <commctrl.h> /* for CommandBar */
#include <aygshell.h> /* for SHMENUBARINFO */
#include <assert.h>
#include <time.h>
#include "resource.h"
#include "game.h"
#include "spider.h"
#include "aboutdlg.h"
#include "optionsdlg.h"

#define MAX_LOADSTRING 100
#define FAN_SPACING 10

/* global variables */
HINSTANCE hInst = NULL;
HWND hwndMain = NULL;
static HWND hwndCB = NULL;       /* the command bar handle */
static HBITMAP hbmpCards = NULL;
static HDC hdcCards = NULL;
static HDC hdcSuitIcons = NULL;

#define MARGIN_LEFT  6
#define MARGIN_TOP   6
#define MARGIN_RIGHT 6
#define BKGROUND_COLOR RGB(0,192,0)

static int selectedColumn = -1;
static int selectedRow = -1;
#define IS_A_CARD_SELECTED() (selectedColumn >= 0 && selectedRow >= 0)

/* function prototypes */
static void drawSuitSymbol(HDC hdc, int suit, int x, int y);
static void getCardDisplayAreaRect(HWND hWnd, int col, int row, RECT *pRect);
static BOOL okToDrop(int startSlot, int endSlot, CardList *cl);
static BOOL okToPickUp(CardList *cl);
static LRESULT onCreate(HWND hWnd, CREATESTRUCT *cs);
static void onGameNewClicked(HWND hWnd);
static void onGameRestartClicked(HWND hWnd);
static void onGameOptionsClicked(HWND hWnd);
static void requestMoveCards(HWND hWnd, int fromCol, int fromRow, int destCol);
static void selectCard(HWND hwnd, int col, int row);
static LRESULT CALLBACK WndProc(HWND hWnd, UINT message,
		WPARAM wParam, LPARAM lParam);

static void
getCardDisplayAreaRect(HWND hWnd, int col, int row, RECT *pRect)
{
	RECT rt;

	assert(pRect != NULL);

	/* determine size of window */
	GetClientRect(hWnd, &rt);
	int screenWidth = rt.right - rt.left;
	int tableauWidth = screenWidth - MARGIN_RIGHT - MARGIN_LEFT;

	pRect->left = MARGIN_LEFT + tableauWidth * col / 10;
	pRect->top = MARGIN_TOP + FAN_SPACING * row;
	pRect->right = pRect->left + CARD_SIZE_X;
	pRect->bottom = pRect->top + CARD_SIZE_Y;
}

/**
 * @return true to continue running the application, false to
 *         exit the program
 */
static BOOL
InitInstance(HINSTANCE hInstance, int nCmdShow)
{
	TCHAR szTitle[MAX_LOADSTRING];
	TCHAR szWindowClass[MAX_LOADSTRING];

	/* initialize global variable */
	hInst = hInstance;

	/* initialize global strings */
	LoadString(hInstance, IDC_APP_WINDOW_CLASS, szWindowClass, MAX_LOADSTRING);
	LoadString(hInstance, IDS_APP_TITLE, szTitle, MAX_LOADSTRING);

	/* if it is already running, then focus on the window */
	hwndMain = FindWindow(szWindowClass, szTitle);
	if (hwndMain)
	{
		SetForegroundWindow((HWND) (((DWORD)hwndMain) | 0x01));
		return FALSE;
	}

	/* load the card images */
	hbmpCards = LoadBitmap(hInst, L"Cards");

	/* seed the random number generator */
	srand(time(NULL));

	/* register the window class */
	HBRUSH hbrBk = CreateSolidBrush(BKGROUND_COLOR);
	WNDCLASS wc = {
		.style = CS_HREDRAW | CS_VREDRAW,
		.lpfnWndProc = (WNDPROC) WndProc,
		.cbClsExtra = 0,
		.cbWndExtra = 0,
		.hInstance = hInstance,
		.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_SPIDER)),
		.hCursor = NULL,
		.hbrBackground = hbrBk,
		.lpszMenuName = NULL,
		.lpszClassName = szWindowClass,
		};
	if (RegisterClass(&wc) == 0)
		return FALSE;

	hwndMain = CreateWindow(szWindowClass, szTitle,
	                    WS_CLIPCHILDREN | WS_VISIBLE,
	                    CW_USEDEFAULT, CW_USEDEFAULT,
	                    CW_USEDEFAULT, CW_USEDEFAULT,
	                    NULL, NULL, hInstance, NULL);
	if (hwndMain == NULL)
		return FALSE;

	/* resize window to accomodate menu bar */
	{
		RECT rc;
		//const int MENU_HEIGHT = 26;
		const int MENU_HEIGHT = 52;

		GetWindowRect(hwndMain, &rc);
		//assert(hwndCB != NULL);
		//RECT rc1;
		//GetWindowRect(hwndCB, &rc1);
		//rc.bottom -= (rc1.bottom - rc1.top);

		rc.bottom -= MENU_HEIGHT;
		//if (hwndCB != NULL)
			MoveWindow(hwndMain, rc.left, rc.top, rc.right, rc.bottom, FALSE);
	}

	game_shuffle();
	start_game();

	ShowWindow(hwndMain, nCmdShow);
	UpdateWindow(hwndMain);

	return TRUE;
}

void
main_invalidate_all(void)
{
	RECT rectScreen;
	GetClientRect(hwndMain, &rectScreen);
	InvalidateRect(hwndMain, &rectScreen, TRUE);
}

void
main_invalidate_card_rect(int slot, int firstCard, int numCards)
{
	RECT rect;
	RECT rectScreen;

	if (firstCard < 0)
	{
		/* negative means count that many cards from the end of the pile */
		firstCard = cardlist_len(tableau[slot]) + firstCard;
	}
	getCardDisplayAreaRect(hwndMain, slot, firstCard, &rect);

	if (numCards == 0)
	{
		/* want the whole column to be repainted */
		GetClientRect(hwndMain, &rectScreen);
		rect.top = rectScreen.top;
		rect.bottom = rectScreen.bottom;
	}

	InvalidateRect(hwndMain, &rect, TRUE);
}

void
main_invalidate_status_area(void)
{
	/* TODO- use a RECT that only covers the text area,
	 * this is just lazy: */
	main_invalidate_all();
}

static BOOL
okToDrop(int startSlot, int endSlot, CardList *cl)
{
	assert(startSlot >= 0 && startSlot < NUM_TABLEAU_SLOTS);
	assert(endSlot >= 0 && endSlot < NUM_TABLEAU_SLOTS);
	assert(cl != NULL);

	if (startSlot == endSlot)
	{
		/* can't drop on self */
		return FALSE;
	}

	CardList *destCL = tableau[endSlot];
	if (destCL == NULL)
	{
		/* always able to drop on an empty slot */
		return TRUE;
	}

	/* find last card in slot */
	while (destCL->next != NULL)
		destCL = destCL->next;

	/* check whether the card can be stacked there */
	return (destCL->card.rank == cl->card.rank + 1);
}

/**
 * Check whether the given pile of cards can be picked up from
 * the tableau as one unit.
 */
static BOOL
okToPickUp(CardList *cl)
{
	if (cl == NULL)
		return FALSE;

	/* face-down cards cannot be moved */
	if (!cl->card.face_up)
		return FALSE;

	/* the top card in a pile can always be moved */
	if (cl->next == NULL)
		return TRUE;

	/* multiple cards can be moved if the suit matches and rank differs
	 * by one */
	if (cl->next->card.suit == cl->card.suit
		&& cl->next->card.rank == cl->card.rank - 1)
		return okToPickUp(cl->next);

	return FALSE;
}

static void
onButtonDown(HWND hWnd, int xPos, int yPos)
{
	RECT rt;

	/* determine size of window */
	GetClientRect(hWnd, &rt);
	int screenWidth = rt.right - rt.left;
	int tableauWidth = screenWidth - MARGIN_RIGHT - MARGIN_LEFT;

	/* really we want x = (xPos - MARGIN_LEFT) / (tableauWidth / 10),
	 * but doing it this way avoids the rounding errors */
	int x = (xPos - MARGIN_LEFT) * 10 / tableauWidth;
	int y = (yPos - MARGIN_TOP) / FAN_SPACING;
	if (x < 0 || x >= 10 || y < 0)
	{
		selectCard(hWnd, -1, -1);
		return;
	}

	/* how many cards are in the clicked slot? */
	int numCardsInSlot = cardlist_len(tableau[x]);

	/* if click is past the end of the last card in the pile... */
	if (yPos > MARGIN_TOP + (numCardsInSlot - 1) * FAN_SPACING + CARD_SIZE_Y)
	{
		selectCard(hWnd, -1, -1);
		return;
	}

	/* selected last card in pile */
	if (y >= numCardsInSlot)
		y = numCardsInSlot - 1;

	if (IS_A_CARD_SELECTED() && x != selectedColumn)
	{
		// attempt move
		requestMoveCards(hWnd, selectedColumn, selectedRow, x);
		selectCard(hWnd, -1, -1);
	}
	else
	{
		selectCard(hWnd, x, y);
	}
}

static LRESULT
onCreate(HWND hWnd, CREATESTRUCT *pcs)
{
	SHMENUBARINFO mbi;

	memset(&mbi, 0, sizeof(SHMENUBARINFO));
	mbi.cbSize = sizeof(SHMENUBARINFO);
	mbi.hwndParent = hWnd;
	mbi.dwFlags = 0;
	mbi.nToolBarId = IDM_MENU;
	mbi.hInstRes = hInst;
	mbi.nBmpId = 0;
	mbi.cBmpImages = 0;

	assert(hInst != NULL);
	assert(hWnd != NULL);
	assert(hwndCB == NULL);

	if (!SHCreateMenuBar(&mbi))
	{
		MessageBox(NULL, L"SHCreateMenuBar failed", L"Debug", MB_OK);
		return -1;
	}

	hwndCB = mbi.hwndMB;
	assert(hwndCB != NULL);
	return 0;
}

static void
onDealClicked(HWND hWnd)
{
	/* check whether any cards left to deal */
	if (remainingCards == 0)
	{
		MessageBox(hWnd, L"No more cards to deal", L"Deal", MB_OK);
		return;
	}

	/* check for empty slots */
	int i = 0;
	BOOL bEmpty = FALSE;
	for (i = 0; i < NUM_TABLEAU_SLOTS; i++)
	{
		if (tableau[i] == NULL)
			bEmpty = TRUE;
	}

	/* if the player has already completed 7 sequences, then
	 * it is impossible to fill in all the empty slots */
	if (bEmpty && game_get_completed_count() < 7)
	{
		MessageBox(hWnd, L"Fill in empty pile first", L"Deal", MB_OK);
		return;
	}

	deal_more_cards();
}

static void
onUndoClicked(HWND hWnd)
{
	if (undo != NULL)
	{
		undo_last_move();

		/* force a redraw of entire screen */
		RECT rt;
		GetClientRect(hWnd, &rt);
		InvalidateRect(hWnd, &rt, TRUE);
	}
	else
	{
		MessageBox(hWnd, L"Nothing to Undo", L"Error", MB_OK);
	}
}

static void
drawCard(HDC hdc, Card card, int x, int y)
{
	int srcRow;
	int srcCol;

	if (card.face_up)
	{
		srcCol = card.rank;
		srcRow = card.suit;
	}
	else
	{
		srcCol = 2;
		srcRow = 4;
	}

	/* paint the card */
	BitBlt(hdc, x, y, CARD_SIZE_X, CARD_SIZE_Y, hdcCards,
		srcCol * CARD_SIZE_X, srcRow * CARD_SIZE_Y, SRCCOPY);
}

static void
drawSuitSymbol(HDC hdc, int suit, int x, int y)
{
	BitBlt(hdc, x, y, SUIT_ICON_SIZE_X, SUIT_ICON_SIZE_Y,
		hdcSuitIcons, suit * SUIT_ICON_SIZE_X, 0, SRCCOPY);
}

static void
highlightCard(HDC hdc, int x, int y)
{
	BitBlt(hdc, x, y, CARD_SIZE_X, CARD_SIZE_Y, hdcCards, 0, 0, DSTINVERT);
}

static void
onGameNewClicked(HWND hWnd)
{
	cleanup_game();
	game_shuffle();
	start_game();
}

static void
onGameRestartClicked(HWND hWnd)
{
	cleanup_game();
	start_game();
}

static void
onGameOptionsClicked(HWND hWnd)
{
	/* popup Options dialog box */
	DialogBox(hInst, (LPCTSTR) IDD_OPTIONS, hWnd, (DLGPROC) OptionsBox);
}

static void
onHelpAboutClicked(HWND hWnd)
{
	/* popup About dialog box */
	DialogBox(hInst, (LPCTSTR) IDD_ABOUT, hWnd, (DLGPROC) AboutBox);
}

static void
onPaint(HWND hWnd)
{
	HDC hdc;
	PAINTSTRUCT ps;
	RECT rt;

	hdc = BeginPaint(hWnd, &ps);

	/* determine size of window */
	GetClientRect(hWnd, &rt);

	/* no longer need to explicitly paint the background
	 * (BeginPaint does that for us) */
	SelectObject(hdc, GetStockObject(WHITE_BRUSH));
	SelectObject(hdc, GetStockObject(BLACK_PEN));

	/* prepare a DC for cards */
	if (hdcCards == NULL)
	{
		hdcCards = CreateCompatibleDC(hdc);
		SelectObject(hdcCards, hbmpCards);
	}
	assert(hdcCards != NULL);
	if (hdcSuitIcons == NULL)
	{
		HBITMAP hbmpSuitIcons = LoadBitmap(hInst, L"Suits");
		hdcSuitIcons = CreateCompatibleDC(hdc);
		SelectObject(hdcSuitIcons, hbmpSuitIcons);
	}
	assert(hdcSuitIcons != NULL);

	/* determine size of cards */
	int screenWidth = rt.right - rt.left;
	int tableauWidth = screenWidth - MARGIN_RIGHT - MARGIN_LEFT;

	/* draw the tableau */
	int i;
	for (i = 0; i < 10; i++)
	{
		CardList *cl;
		int x = MARGIN_LEFT + tableauWidth * i / 10;
		int y = MARGIN_TOP;

		cl = tableau[i];
		while (cl != NULL)
		{
			drawCard(hdc, cl->card, x, y);
			if (selectedColumn == i && cl->next == NULL)
				highlightCard(hdc, x, y);
			y += FAN_SPACING;
			cl = cl->next;
		}
	}

	/* write status text */
	rt.left += 1;
	rt.bottom -= 1;
	rt.right -= 1;
	TCHAR buf[100];
	wsprintf(buf, L"Remaining: %d", remainingCards);
	SetTextColor(hdc, RGB(255,255,255));
	SetBkColor(hdc, BKGROUND_COLOR);
	DrawText(hdc, buf, wcslen(buf),
	         &rt, DT_BOTTOM | DT_LEFT);

	BOOL anyDone = FALSE;
	for (i = 0; i < 8; i++)
	{
		if (finished_piles[i])
		{
			drawSuitSymbol(hdc, get_suit_of_finished_pile(i),
				rt.right - SUIT_ICON_SIZE_X - 2,
				rt.bottom - SUIT_ICON_SIZE_Y);
			rt.right -= (SUIT_ICON_SIZE_X + 2);
			anyDone = TRUE;
		}
	}

	wsprintf(buf, anyDone ? L"Done:" : L"Done: (none)");
	DrawText(hdc, buf, wcslen(buf),
			&rt, DT_BOTTOM | DT_RIGHT);

	/* finished painting */
	EndPaint(hWnd, &ps);
}

static void
requestMoveCards(HWND hWnd, int fromCol, int fromRow, int destCol)
{
	CardList *srcCL;

	fromRow = 0;
	while ((srcCL = getCardListAtPosition(fromCol, fromRow)) != NULL)
	{
		if (okToPickUp(srcCL) && okToDrop(fromCol, destCol, srcCL))
		{
			moveCards(fromCol, fromRow, destCol, TRUE);
			return;
		}
		fromRow++;
	}

	MessageBox(hWnd, L"Can't move those cards", L"Error", MB_OK);
	return;
}

static void
selectCard(HWND hWnd, int col, int row)
{
	if (IS_A_CARD_SELECTED())
		main_invalidate_card_rect(selectedColumn, -1, 1);

	selectedColumn = col;
	selectedRow = row;

	if (IS_A_CARD_SELECTED())
		main_invalidate_card_rect(selectedColumn, -1, 1);
}

static LRESULT CALLBACK
WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDM_GAME_NEW:
			onGameNewClicked(hWnd);
			break;
		case IDM_GAME_RESTART:
			onGameRestartClicked(hWnd);
			break;
		case IDM_GAME_OPTIONS:
			onGameOptionsClicked(hWnd);
			break;
		case IDM_GAME_EXIT:
			SendMessage(hWnd, WM_ACTIVATE, MAKEWPARAM(WA_INACTIVE, 0), (LPARAM) hWnd);
			SendMessage(hWnd, WM_CLOSE, 0, 0);
			break;
		case IDM_EDIT_DEAL:
			onDealClicked(hWnd);
			break;
		case IDM_EDIT_UNDO:
			onUndoClicked(hWnd);
			break;
		case IDM_HELP_ABOUT:
			onHelpAboutClicked(hWnd);
			break;
		default:
			return DefWindowProc(hWnd, message, wParam, lParam);
		}
		break;

	case WM_CREATE:
		return onCreate(hWnd, (CREATESTRUCT *) lParam);

	case WM_LBUTTONDOWN:
		onButtonDown(hWnd, LOWORD(lParam), HIWORD(lParam));
		return 0;
	case WM_PAINT:
		onPaint(hWnd);
		break;
	case WM_DESTROY:
		PostQuitMessage(0);
		break;
	default:
		return DefWindowProc(hWnd, message, wParam, lParam);
	}
	return 0;
}

int WINAPI
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
        LPTSTR lpCmdLine, int nCmdShow)
{
	MSG msg;

	/* perform application initialization */
	if (!InitInstance(hInstance, nCmdShow))
		return FALSE;

	/* main message loop */
	while (GetMessage(&msg, NULL, 0, 0))
	{
		DispatchMessage(&msg);
	}
	return msg.wParam;
}
